
class Widgets:
    # Initialize the class with empty dictionaries for each widget name
    def __init__(self):
        self.buttons = {}
        self.canvases = {}
        self.checkbuttons = {}
        self.entries = {}
        self.frames = {}
        self.labels = {}
        self.labelframes = {}
        self.listboxes = {}
        self.menus = {}
        self.menubuttons = {}
        self.messages = {}
        self.optionmenus = {}
        self.panedwindows = {}
        self.radiobuttons = {}
        self.scales = {}
        self.scrollbars = {}
        self.texts = {}
        self.toplevels = {}


class WidgetOrganizer:
    def __init__(self) -> None:
        self._widgets = Widgets()

    @property
    def widget(self):
        return self._widgets


import pandas as pd

available_file_types = {
    0   :   [".tab", "table", "tabular"],
    1   :   [".csv","csv"],
    2   :   [".json","json"],
    3   :   [".xls", "xls"],
}

def setup(settings):
    """
    Creates and returns a PSTLDataFrame object based on settings dictionary passed in.
    The settings parameter must have keys 'file' and 'file_type'.
    
    Keys:
        'file'      : str   ->  name of plasma ['cylinderical', 'spherical', or 'planar']
        'file_type' : str   ->  type of data file to read in ['.csv', '.xls', or '.json']
    (optional)
        'name'          : str   ->  name designation for data object
        'description'   : str   ->  description of data object
        'args'          : tuple ->  addional position arguments
        'kwargs'        : dict  ->  addional keyword arguments
    Returns: PSTLDataFrame Object

    Other Notes:
        common kwargs for 'file_type' : ".csv"
            "sep" --or-- "delimiter"    :   str         ->  tabular table delimiter i.e. "," --or-- "\t"
            "header"                    :   int | None  ->  location of header  
                i.e. 
                    if no-header then None
                    if first line header then 0
                    if names is given then header should be 0
            "names"     :   iterable(str,...)           ->  column names i.e. ["voltage", "current"]
            "skiprows"  :   int | list[int,...] | None  ->  # of rows or list of rows to skip
    """
    shape = settings["file_type"].lower()
    if shape in available_file_types[0]:
        get_data_func = pd.read_table
    elif shape in available_file_types[1]:
        get_data_func = pd.read_csv
    elif shape in available_file_types[2]:
        get_data_func = pd.read_json
    elif shape in available_file_types[3]:
        get_data_func = pd.read_excel
    else:
        raise ValueError("'%s' is not a valid option."%(shape))
    negative = settings.get("negative", False)
    name = settings.get("name",None)
    description = settings.get("name",None)
    file = settings["file"]
    args = settings.get("args", (None))
    kwargs = settings.get("kwargs",{})
    dataframe_args = kwargs.get("dataframe_args", [])
    dataframe_kwargs = kwargs.get("dataframe_kwargs",{})
    data = get_data_func(file,*args,**kwargs)
    data.iloc[:,1] = data.iloc[:,1]*-1 if negative is True else data.iloc[:,1] # type: ignore
    pstl_data = PSTLDataFrame(data,*dataframe_args,
                              name=name,description=description,
                              **dataframe_kwargs) # type: ignore
    return pstl_data

class PSTLDataFrame(pd.DataFrame):
    """
    Built on top of a pandas DataFrame, this class adds attributes 'name' and 'description'
    simply for tracking and identifying dataframes more easily
    """
    def __init__(self,*args,name=None,description=None,**kwargs) -> None:
        super().__init__(*args,**kwargs)
        self._name = name
        self._description = description

    @property
    def name(self):
        return self._name
    @name.setter
    def name(self, string):
        if isinstance(string, (str, type(None))):
            self._name = string
        else:
            raise TypeError("'%s' Must be a str or None type, not %s"%(str(string),str(type(string))))
    @property
    def description(self):
        return self._description
    @description.setter
    def description(self, string):
        if isinstance(string, (str, type(None))):
            self._description = string
        else:
            raise TypeError("Description change must be str or None type, not type '%s'"%(str(type(string))))




from .helpers import (shape_fit_selector, shape_func_selector)
from .shape_fit_methods import (
    cylinderical_fit_method,
    spherical_fit_method,
    planar_fit_method
)
from .shape_func_methods import (
    cylinderical_func_method,
    spherical_domain_condition,
    planar_func_method,
)

from .helpers import (
    available_methods,
    available_fit_functions,
    available_func_functions,
    fit_options,
    func_options,
)


"""
For the thin sheath case there are two methods for determining ion current
    1) Fit Method: Emperically fits a line or function to the data
    2) Function Method: Theoritical function results
"""


import tkinter as tk

from PIL import Image, ImageTk

from pstl.extras.images import pstl_ico_path, pstl_png_path


def main():
    root = tk.Tk()
    img_16 = tk.PhotoImage(file="/home/tyjoto/temp/tkinter-icons/icon-16.png")
    img_32 = tk.PhotoImage(file="/home/tyjoto/temp/tkinter-icons/icon-32.png")
    #img = tk.PhotoImage(file=pstl_ico_path)
    #img = tk.PhotoImage(file=pstl_png_path)
    #root.tk.call('wm', 'iconphoto', root._w, img)

    img = Image.open(pstl_ico_path)
    img.save('/home/tyjoto/github-work/pstl-tools/src/pstl/extras/images/pstl-16.ico',format = 'ICO', sizes=[(16,16)])
    img.save('/home/tyjoto/github-work/pstl-tools/src/pstl/extras/images/pstl-32.ico',format = 'ICO', sizes=[(32,32)])
    tkimage_16 = ImageTk.PhotoImage(Image.open("/home/tyjoto/github-work/pstl-tools/src/pstl/extras/images/pstl-16.ico"))
    tkimage_32 = ImageTk.PhotoImage(Image.open("/home/tyjoto/github-work/pstl-tools/src/pstl/extras/images/pstl-32.ico"))
    #label=tk.Label(root,image=tkimage)

    #label.pack()
    root.iconphoto(False,tkimage_32, img_16)
    root.mainloop()

if __name__ == "__main__":
    main()


from abc import ABC, abstractmethod, abstractproperty
from typing import Dict, Callable, Any
import json
import pprint

import numpy as np
import pandas as pd

from pstl.utls.abc import PSTLObject
from pstl.utls.json import load_build_and_get_parameters
from pstl.utls.plasmas import Plasma
from pstl.utls.plasmas import setup as plasma_setup
from pstl.utls.data import setup as data_setup
from pstl.utls.objects import setup as object_setup
from pstl.diagnostics.probes.langmuir.single import setup as probe_setup
from pstl.diagnostics.probes import Probe

from pstl.utls.preprocessing import preprocessing_filter, preprocess_filter, smooth_filter
from pstl.diagnostics.probes.langmuir.single import SingleProbeLangmuir

from pstl.diagnostics.probes.langmuir.single.analysis.algorithms import topham, lobbia
from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import get_floating_potential
from pstl.diagnostics.probes.langmuir.single.analysis.ion_saturation_current import get_ion_saturation_current, get_ion_saturation_current_density
from pstl.diagnostics.probes.langmuir.single.analysis.electron_temperaure import get_electron_temperature
from pstl.diagnostics.probes.langmuir.single.analysis.electron_saturation_current import get_electron_saturation_current, get_electron_saturation_current_density
from pstl.diagnostics.probes.langmuir.single.analysis.plasma_potential import get_plasma_potential
from pstl.diagnostics.probes.langmuir.single.analysis.electron_density import get_electron_density
from pstl.diagnostics.probes.langmuir.single.analysis.ion_density import get_ion_density
from pstl.utls.constants import lambda_D
from pstl.utls.helpers import probe_radius_to_debye_length as get_sheath_type
from pstl.utls.helpers import method_function_combiner, method_selector
from pstl.utls.decorators import absorb_extra_args_kwargs, add_empty_dict
get_lambda_D = add_empty_dict(absorb_extra_args_kwargs(lambda_D))
get_sheath_type = add_empty_dict(absorb_extra_args_kwargs(get_sheath_type))

# maybe a concrete class called ProbeData that would jsut be raw_data, deleted_data, filtered_data, smoothed_data
singe_langmuir_probe_solver_setup_builders = {
    "plasma"    :   plasma_setup,
    "probe"     :   probe_setup,
    "data"      :   data_setup,
}


def setup(settings, *args, **kwargs):
    """
    Creates and returns a SingleProbeLangmuirSolver object based on settings dictionary passed in.
    The settings parameter must have keys 'plasma', 'probe', and 'data', where in
    each is a dictionary with either a key being 'define' or 'file_load'. If 'define',
    there is another dictionary that defines all need properties to make the object. In
    the 'file_load' case, the entry is a string that is the file path to load a json file
    that has all the needed properties to make the object.
    
    Keys (mandatory):
        'plasma': dict[dict | str]   ->  Plasma object defining properties or path to json file 
        'probe' : dict[dict | str]   ->  Probe object defining properties or path to json file
        'data'  : dict[dict | str]   ->  Data object defining properties or path to json file
    (optional)
        'name'          : str   ->  name designation for object
        'description'   : str   ->  description of object
        'args'          : tuple ->  addional position arguments
        'kwargs'        : dict  ->  addional keyword arguments
    Returns: Solver Object
    """
    # create new object with parameters (arguments)
    output_object: SingleLangmuirProbeSolver = object_setup(
        *args,
        settings=settings,
        builders=singe_langmuir_probe_solver_setup_builders,
        Builder=SingleLangmuirProbeSolver,
        **kwargs,
    )

    return output_object


class DiagnosticData(PSTLObject):
    def __init__(self, data, deleted_data=None, filtered_data=None, smoothed_data=None, *args, **kwargs):
        # set different data versions
        self.set_data(
            data,
            deleted_data=deleted_data,
            filtered_data=filtered_data,
            smoothed_data=smoothed_data,
        )

    def set_data(self, raw_data, source=0, deleted_data=None, filtered_data=None, smoothed_data=None):
        """
        Takes raw_data
        """
        self._raw_data = raw_data
        self._deleted_data = deleted_data
        self._filtered_data = filtered_data
        self._smoothed_data = smoothed_data

        # available sources
        available_sources = {
            0: "best",
            1: "raw_data",
            2: "filtered_data",
            3: "smoothed_data",
        }

        # Converts method: str -> method: int if method is a str
        if isinstance(source, str):
            reversed_sources = {v: k for k, v in available_sources.items()}
            source = reversed_sources.get(source, None)

        # choose data source
        if source == 0: # Best
            if smoothed_data is True:
                # get smoothed of filtered data
                raise NotImplementedError
                data = smoothed_data
            elif isinstance(smoothed_data, pd.DataFrame):
                data = smoothed_data
            elif filtered_data is True:
                # get filtered data
                raise NotImplementedError
                data = filtered_data
            elif isinstance(filtered_data, pd.DataFrame):
                data = filtered_data
            else:
                data = raw_data

        elif source == 1:   # raw data
            data = raw_data

        elif source == 2:   # filtered data
            if filtered_data is True:
                filtered_data = pd.DataFrame
            elif filtered_data is False or filtered_data is None:
                filtered_data = raw_data
            elif isinstance(filtered_data, pd.DataFrame):
                filtered_data = filtered_data
            else:
                raise TypeError(
                    "'filtered_data' can only type pd.DataFrame, bool, None: ", type(filtered_data))
            data = filtered_data

        elif source == 3:   # tecniquely filtered and and smoothed if filtered is True and smoothed is True
            if smoothed_data is True:
                smoothed_data = pd.DataFrame
            elif smoothed_data is False or smoothed_data is None:
                smoothed_data = raw_data
            elif isinstance(smoothed_data, pd.DataFrame):
                smoothed_data = smoothed_data
            else:
                raise TypeError(
                    "'smoothed_data' can only type pd.DataFrame, bool, None: ", type(smoothed_data))
            data = smoothed_data

        else:  # makes a table of options if error occurs
            table = "\n".join(
                [f"{k}\t{v}" for k, v in available_sources.items()])
            raise ValueError(
                f"Matching source not found: {source}\nChoose from one of the available options:\n{table}")

        # based on data argument set data variable to use
        self._data = data

    def preprocess(self, *args, source=0, delete=True, **kwargs):
        filtered_data, deleted_data = preprocess_filter(
            self._data, *args, delete=delete, **kwargs)
        self.set_data(self._data, source=source,
                      deleted_data=deleted_data, filtered_data=filtered_data)

    def smooth(self, *args, source=0, **kwargs):
        smoothed_data = smooth_filter(self._data, *args, **kwargs)
        self.set_data(self._data, source=source, smoothed_data=smoothed_data)


class PlasmaProbeSolver(DiagnosticData, ABC):
    def __init__(self, Plasma: Plasma, Probe, Data: pd.DataFrame,
                 methods: Dict = {}, properties: Dict = {},
                 deleted_data: pd.DataFrame | bool | None = None,
                 filtered_data: pd.DataFrame | bool | None = None,
                 smoothed_data: pd.DataFrame | bool | None = None,
                 name: str | None = None, description: str | None = None,
                 *args, **kwargs) -> None:
        # super(ABC, self).__init__()
        DiagnosticData.__init__(
            self, Data,
            deleted_data=deleted_data, filtered_data=filtered_data, smoothed_data=smoothed_data)

        # add verification here
        self._plasma = Plasma
        self._probe = Probe

        # create results dictionary
        self._results = self.set_available_plasma_properties(properties)

        # set defeult methods (abstract)
        self._methods = self.set_default_methods(methods)

        # set defeult methods (abstract)
        self._methods_args = self.set_default_methods_args(kwargs)
        
        # set defeult methods (abstract)
        self._methods_kwargs = self.set_default_methods_kwargs(kwargs)
    @property
    def plasma(self):
        return self._plasma

    @property
    def probe(self):
        return self._probe

    @property
    def data(self):
        return self._data

    @property
    def raw_data(self):
        return self._raw_data

    @property
    def deleted_data(self):
        return self._deleted_data

    @property
    def results(self):
        return self._results

    @property
    def methods(self):
        return self._methods

    @property
    def methods_args(self):
        return self._methods_args
    
    @property
    def methods_kwargs(self):
        return self._methods_kwargs

    @abstractmethod
    def set_default_methods(self, methods: dict) -> dict:
        pass
        # return {}

    @abstractmethod
    def set_default_methods_args(self, methods: dict) -> tuple:
        pass

    @abstractmethod
    def set_default_methods_kwargs(self, methods: dict) -> dict:
        pass

    @abstractmethod
    def set_available_plasma_properties(self, properties: dict) -> dict:
        pass

    @abstractmethod
    def solve(self, *args, methods: dict={}, **kwargs):
        pass


class PlasmaLangmuirProbeSolver(PlasmaProbeSolver):
    def __init__(self, Plasma: Plasma, Probe, Data, *args, **kwargs) -> None:
        super().__init__(Plasma, Probe, Data, *args, **kwargs)

        self._data_e = self.data.copy() # type: ignore

    def update_current_e(self, curret_i):
        self.data_e.current = np.subtract(self.data.current, curret_i)  # type: ignore

    @property
    def data_e(self):
        return self._data_e


class SingleLangmuirProbeSolver(PlasmaLangmuirProbeSolver):
    def __init__(self, plasma: Plasma, probe: SingleProbeLangmuir, data: pd.DataFrame,
                 methods: Dict = {}, properties: Dict = {},
                 *args, **kwargs) -> None:
        super().__init__(plasma, probe, data,
                         methods=methods, properties=properties,
                         *args, **kwargs)


    def set_available_plasma_properties(self, properties: dict) -> dict:
        super().set_available_plasma_properties(properties)
        available_plasma_properties = {
            "V_f": {'value': None, 'other': None},
            "V_s": {'value': None, 'other': None},
            "KT_e": {'value': None, 'other': None},
            "n_e": {'value': None, 'other': None},
            "I_es": {'value': None, 'other': None},
            "J_es": {'value': None, 'other': None},
            "n_i": {'value': None, 'other': None},
            "I_is": {'value': None, 'other': None},
            "J_is": {'value': None, 'other': None},
            "lambda_De": {'value': None, 'other': None},
            "sheath": {'value': None, 'other': None},
        }
        return available_plasma_properties

    def set_default_methods(self, methods: Dict):
        # methods = super().setdefault_methods(methods)
        super().set_default_methods(methods)
        default_methods = {
            "algorithm" : 0,
        }
        default_methods.update(methods)
        return default_methods
    
    def set_default_methods_args(self, kwargs: dict) -> dict:
        super().set_default_methods_args(kwargs)
        return kwargs.pop("algorithm_args",())
    
    def set_default_methods_kwargs(self, kwargs: dict) -> dict:
        super().set_default_methods_kwargs(kwargs)
        return kwargs.pop("algorithm_kwargs",{})

    def solve(self, *args,methods={}, **kwargs):
        super().solve(methods,*args,**kwargs)
        # algo is the default algo to use i.e. topham, lobia, etc.
        # verify methods is a dictionary
        # methods should not be modified unless you want to modify which methods are being used during the alogrythm.
        if not isinstance(methods, dict):
            raise ValueError(
                "'methods' must be a dictionary not: ", type(methods))

        # overwrite methods if passed in
        methods_to_use = dict(self.methods)
        methods_to_use.update(methods)

        # grab algorithm args and kwargs from kwargs
        new_algorithm_args = kwargs.get("algorithm_args",tuple())
        new_algorithm_kwargs = kwargs.get("algorithm_kwargs",{})
        

        # update defaults with new
        algorithm_args = new_algorithm_args
        algorithm_kwargs = self.methods_kwargs | new_algorithm_kwargs

        # Choose which alogrithm to solve for results
        algorithm_func = method_selector(methods_to_use["algorithm"], 
                               available_algorithms,
                               available_algorithm_functions,
        )

        # run selected algorithm
        data, results = algorithm_func(   
            self.data.voltage, self.data.current,   # type: ignore
            self.probe.shape, self.probe.radius, self.probe.area,
            self.plasma.m_i, 
            *algorithm_args,
            m_e=self.plasma.m_e,
            **algorithm_kwargs,
        )

        # save results and data to solver object
        self._results = results
        self._data = data

# Declare available methods
available_algorithms = {
    0: ['topham'],
    1: ['lobbia'],
}
# Declare correspondin functions for available_methods
available_algorithm_functions = {
    0: topham,
    1: lobbia,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(
    available_algorithms, available_algorithm_functions)



from pstl.utls.constants import lambda_D as get_lambda_D
from pstl.utls.decorators import absorb_extra_args_kwargs, add_empty_dict
get_lambda_D = add_empty_dict(absorb_extra_args_kwargs(get_lambda_D))
from pstl.utls.plasmas.sheath import get_probe_to_sheath_ratio
from ..ion_density import get_ion_density
from ..electron_density import get_electron_density
from ..plasma_potential import get_plasma_potential
from ..electron_saturation_current import get_electron_saturation_current, get_electron_saturation_current_density
from ..electron_temperaure import get_electron_temperature
from ..ion_saturation_current import get_ion_saturation_current, get_ion_saturation_current_density
from ..floating_potential import get_floating_potential
from ..ion_current import get_ion_current

default_methods_keys = {
    "V_f":              ["V_f",             "floating_potential"],
    "V_s":              ["V_s",             "plasma_potential", "space_potential"],
    "KT_e":             ["KT_e",            "electron_temperature", "T_e", "T_eV"],
    "n_e":              ["n_e",             "electron_density"],
    "n_i":              ["n_i",             "ion_density"],
    "I_es":             ["I_es",            "electron_saturation_current"],
    "I_is":             ["I_is",            "ion_saturation_current"],
    "J_es":             ["J_es",            "electron_saturation_current_density"],
    "J_is":             ["J_is",            "ion_saturation_current_density"],
    "lambda_De":        ["lambda_De",       "electron_debye_length"],
    "r_p/lambda_De":    ["r_p/lambda_De",   "ratio"],
    "sheath":           ["sheath",          "sheath_type"],
    "I_i_thin":         ["I_i_thin",        "ion_current_thin_sheath"],
    "I_i_transitional": ["I_i_transitional","ion_current_transitional_sheath"],
    "I_i_thick":        ["I_i_thick",       "ion_current_thick_sheath"],
}

default_plasma_properties_to_get = [
    "V_f",
    "V_s",
    "KT_e",
    "n_e",
    "n_i",
    "I_es",
    "J_es",
    "I_is",
    "J_is",
    "lambda_De",
    "r_p/lambda_De",
    "sheath",
]
default_methods = {
    "V_f": "consecutive",
    "V_s": "intersection",
    "KT_e": 0,
    "n_e": 0,
    "n_i": 0,
    "I_es": 0,
    "I_is": 3,
    "J_es": 0,
    "J_is": 0,
    "lambda_De": 0,
    "r_p/lambda_De": 0,
    "sheath": 0,
}


def print_results(results):
    for key, value in results.items():
        print(f"{key}:")
        for inner_key, inner_value in value.items():
            if key in ["n_i", "n_e", "I_is", "I_es", "J_es", "J_is", "lambda_De"] and inner_key == "value":
                if inner_value is not None:
                    print(f"\t{inner_key}: {inner_value:.2e}")
                else:
                    print(f"\t{inner_key}: {inner_value}")
            elif key in ["V_f", "V_s", "KT_e"] and inner_key == "value":
                if inner_value is not None:
                    print(f"\t{inner_key}: {inner_value:.2f}")
                else:
                    print(f"\t{inner_key}: {inner_value}")
            else:
                print(f"\t{inner_key}: {inner_value}")


def _return_orgainizer(returns):
    # check if tuple for indexing
    if isinstance(returns, tuple):
        value = returns[0]
        other = returns[1:]
    else:
        value = returns
        other = None
    return value, other


def get_plasma_property_sorter(key, method, *args, **kwargs):

    # choose which property
    if key == "V_f":
        func = get_floating_potential
    elif key == "V_s":
        func = get_plasma_potential
    elif key == "I_is":
        func = get_ion_saturation_current
    elif key == "n_i":
        func = get_ion_density
    elif key == "I_es":
        func = get_electron_saturation_current
    elif key == "KT_e":
        func = get_electron_temperature
    elif key == "n_e":
        func = get_electron_density
    elif key == "J_es":
        func = get_electron_saturation_current_density
    elif key == "J_is":
        func = get_ion_saturation_current_density
    elif key == "lambda_De":
        func = get_lambda_D
    elif key == "sheath":  # key == "r_p/lamda_De":
        func = get_probe_to_sheath_ratio
    else:
        table = "\n".join(
            [f"{k}\t{v}" for k, v in enumerate(default_methods)])
        raise ValueError(
            f"Matching key not found: {key}\nChoose from one of the available options:\n{table}")

    # solve and return a tuple
    return func(*args, method=method, **kwargs)

def topham_configure(convergence_percent, methods, *args, **kwargs):
    """
    This function calls other functions and then returns all of the 
    the initial configurations for the topham algorithm for solving a
    Langmuir probe.

    Parameters:
        convergence_percent     "The convergence criteria in percent of the
                                Langmuir algorithm for changes in sheath"
        methods                 "Dictionary of methods for solving for 
                                plasma properties"
        *args
        **kwargs                "Function args/kwargs for methods"

    Returns:
        convergence_decimal : float
        methods_to_use      : dict[str | int, ...]
        properties_to_get   : list[str, ...]
        results             : dict[float,dict[...]]
        func_args           : dict[tuple, ...]
        func_kwargs         : dict[dict, ...]
    """
    # Initialized plasma properties to return (not yet implemented)
    properties = None

    # Checks defined methods are in a dictionary
    if not isinstance(methods, dict):
        raise ValueError(
            "'methods' must be a dictionary not: ", type(methods))
    
    # convert convergence percent to a decimal
    convergence_decimal = convergence_percent/100

    # overwrite default methods to be used if methods is partially/fully defined
    methods_to_use = dict(default_methods)
    methods_to_use.update(methods)

    # overwrite properties if passed in (not implemented yet)
    if properties is None:
        properties_to_get = list(default_plasma_properties_to_get)
    elif isinstance(properties, list):
        properties_to_get = list(properties)
    else:
        raise ValueError(
            "'properteies' must be a list or None not: ", type(methods))

    # set up results dictionary for returns
    results = {}
    for plasma_property in properties_to_get:
        results[plasma_property] = {'value': None, 'other': None}

    # see what args are given and create a new dictionary
    func_args = {}
    #for key in methods_to_use.keys():
    for key in default_methods_keys.keys():
        #func_args[key] = kwargs.get(key+"_args", ())
        func_args[key] = check_for_args(key,kwargs)
    # see what kwargs are given and create a new dictionary
    func_kwargs = {}
    #for key in methods_to_use.keys():
    for key in default_methods_keys.keys():
        #func_kwargs[key] = kwargs.get(key+"_kwargs", {})
        func_kwargs[key] = check_for_kwargs(key,kwargs)

    # tuple of returns
    returns = (
        convergence_decimal,
        methods_to_use,
        properties_to_get,
        results,
        func_args,
        func_kwargs,
    )

    return returns

def check_for_args(key,kwargs):
    output = dict(kwargs)
    value = ()
    for alt_key in default_methods_keys[key]:
        alt_key_args = alt_key+"_args"
        if alt_key_args in kwargs:
            value = output.pop(alt_key_args)
            #output[key] = value
        
    return value
def check_for_kwargs(key,kwargs):
    output = dict(kwargs)
    value = {}
    returns = {key+"_kwargs":{}}
    for alt_key in default_methods_keys[key]:
        alt_key_kwargs = alt_key+"_kwargs"
        if alt_key_kwargs in kwargs:
            value[alt_key_kwargs] = output.pop(alt_key_kwargs)
            #output[key] = value
            #print("OUTPUT", output)

    # combine multiple givens to one, keeping first alt_key_kwargs as priority
    for alt_key_kwargs, values in reversed(value.items()):
        returns[key+"_kwargs"].update(values)
        
    #print("FINALOUTPUT", output)
    return returns[key+"_kwargs"]
    
def get_iteration_args_kwargs(iteration: int | str, key:str, args: tuple, kwargs: dict) -> tuple[tuple,dict]:
    # initialize this iterations kwargs
    iteration_kwargs: dict = {}
    # rasie value error for 
    def raise_value_error(key):
        raise ValueError("'%s' is not a known key in %s"%(key, str(default_methods_keys[key])))
    # all iteration list
    all_iteration_list = ["all",0,"0"]
    # gets the generalized for all interations kwargs and then updates for the specified interation
    if any(k in kwargs for k in all_iteration_list):
        iteration_kwargs.update(kwargs.get("all", {}))
    if ((iteration in kwargs) or (str(iteration) in kwargs)) and (iteration not in all_iteration_list):
        iteration_kwargs.update(kwargs.get(iteration,{}))

    # set function key for args and kwargs
    func_key = default_methods_keys[key][0] if key in default_methods_keys[key] else raise_value_error(key)
    func_key_args = func_key+"_args"
    func_key_kwargs = func_key+"_kwargs"
    # get function args and kwargs using key value
    func_args: tuple = iteration_kwargs.get(func_key_args, ())
    func_kwargs: dict = iteration_kwargs.get(func_key_kwargs, {})
    return func_args, func_kwargs


def topham_get_ion_current_and_density(
        voltage, current, shape, sheath_method, area, m_i,KT_e, 
        r_p, lambda_De, V_f, V_s, n_e, n_i,*args, **kwargs):
    """
    NEEDS WORK, but it works
    Only for transitional and thick sheaths
    """
    # grab ion density args and kwargs
    ion_density_args = kwargs.get("ion_density_args",())
    ion_density_kwargs = kwargs.get("ion_density_kwargs",{})
    # get ion denisty
    n_i, n_i_extras = get_ion_density(
        voltage, current, 
        *ion_density_args, 
        area=area, m_i=m_i, KT_e=KT_e, shape=shape, r_p=r_p, 
        lambda_De=lambda_De, method=sheath_method,
        **ion_density_kwargs)

    # grab ion current args and kwargs
    ion_current_args = kwargs.get("ion_current_args",())
    ion_current_kwargs = kwargs.get("ion_current_kwargs",{})
    # get new ion current with transitional or thick sheath
    I_i, I_i_extras = get_ion_current(
        shape,
        voltage,
        *ion_current_args,
        current=current,
        method=sheath_method, V_f=V_f, n_i=n_i, n0=n_e, m_i=m_i, KT_e=KT_e, 
        V_s=V_s, area=area,r_p=r_p, 
        lambda_De=lambda_De,
        **ion_current_kwargs,
    )
    return I_i, I_i_extras, n_i, n_i_extras



import numpy as np

# domain conditions
def cylinderical_domain_condition(voltage, V_s, KT_e, *args, **kwargs):
    return _domain_condition(voltage, V_s, KT_e, *args, **kwargs)


def spherical_domain_condition(voltage, V_s, KT_e, *args, **kwargs):
    return _domain_condition(voltage, V_s, KT_e, *args, **kwargs)


def planar_domain_condition(voltage, V_s, KT_e, *args, **kwargs):
    return _domain_condition(voltage, V_s, KT_e, *args, **kwargs)

# generalized domain


def _domain_condition(voltage, V_s, KT_e, *args, **kwargs):
    condition = np.divide(
        np.subtract(V_s, voltage),  # [V]
        KT_e                         # [eV]
    )
    return condition > 1  # >>1


from pstl.gui.langmuir.langmuir import *
from pstl.gui.langmuir import langmuir
from pstl.gui.langmuir import results
from pstl.gui.langmuir import canvas
from pstl.gui.langmuir.solver import gui_langmuir



import numpy as np


from .methods import get_ion_current




def check_for_ion_current_fit(I_i_fit, voltage, current, shape=None, method=None, **kwargs):
    """
    Determines if there is function fit for ion current (I_i_fit). If it does not exist (i.e. I_i_fit is None),
    then a fit is made using suppied voltage, current, shape=None, method=None, kwargs['fit_kwargs'] to get_ion_current function.
    Returns a function fit class."""
    # determine starting point (all positive after vf) if not given
    if I_i_fit is None:
        # determine starting point (all positive after V_f)
        fit_kwargs = kwargs.pop('fit_kwargs', {})
        # get floating potential
        I_i, extras = get_ion_current(
            shape, voltage, current, method=method, fit_kwargs=fit_kwargs,
        )
        I_i_fit = extras["fit"]

    return I_i_fit






import json
from typing import Any, Callable, Iterable


def load_json(file,*args, **kwargs):
    """
    wrapper to with open() as f: format for JSON files such that it is one call.
    Returns a dictionary of JSON data. 
    """
    with open(file) as f:
        json_data = json.load(f)
    return json_data


def load_build_and_get_parameters(
        settings:dict[Any, Any], 
        builders: dict[str,Callable],
        *args, **kwargs) -> Any:
    """
    The 'settings' argument contains the dictionary of the imported JSON
    now check for sub-dictionaries that are length=1 and posses either 
    key words "BUILD", "BUILD_FROM_FILE", or "FROM_FILE".

    If "BUILD" or "BUILD_FROM_FILE":
        then takes either the following dictionary or read-in dictionary 
        from JSON file, respectivly, and creates the object using the 
        matching builder. The builders are limited to known objects that
        are required for the setup file that this function is being called
        to make. Thus, 'builders' arguments are passed into this function 
        from the previous setup function. 
    IF "FROM_FILE":
        then the dictionary from the file is placed under the keyword 
        requesting the file be brought in
    
    Returns:
        output  : dict[Any] -> A dictionary of parameters (arguments) 
        using the keywords defined from the input argument 'settings' of 
        this function as the keywords of the returned dictionary with the
        values now of this new dictionary being either a sub-dictionary(ies)
       (if "FROM_FILE") or an object (if "BUILD" or "BUILD_FROM_FILE")
        
    Note:  
        The retured object of this function is no longer a dictionary of 
        settings, but a dictionary of parameters (arguments) for the 
        setup function to pass in as keyword arguments to initiate the
        object.

    """
    # first step is loop through dictionary layer by layer looking for a
    # pair of a key with a value element that is a dictionary of length=1
    # Once found, send that value element that is a dictionay to <func>
    # That function will either build and return an object --or-- iterate
    # through sub-layers till all "FROM_FILE"s have been resolved and a
    # complete dictionary can be returned.
            
    # key words to look for
    BUILD = "BUILD"
    BUILD_FROM_FILE = "BUILD_FROM_FILE"
    FROM_FILE = "FROM_FILE"
    # create raise key error function
    def key_error(key):
        raise KeyError("'%s' is not a valid key in builders"%(key))
    
    # loop through settings
    parameters: dict[str, Any] = dict(settings)
    for key in settings:
        # Check if dictionary and length of dictionary is 1
        subsettings = settings[key]
        isdict = isinstance(subsettings, dict)
        # Check if length if dictionary is 1 if subsettings is dictionary
        isdictlen1 = (True if len(subsettings) == 1 else False) if isdict is True else False
        # Check if passed and move on if true
        if isdictlen1 is True:
            # get subkey to check
            subkey = [*settings[key]][0]
            # If "BUILD", then check if key in builders, then build 
            if subkey == BUILD:
                subsettings = subsettings[BUILD]
                parameters[key] = builders[key](subsettings) if key in builders else key_error(key) 
            # if "BUILD_FROM_FILE", then check if key in builders, read in JSON, then build 
            elif subkey == BUILD_FROM_FILE:
                # here subsettings is a dictionary of length=1 with only a key being "BUILD_FROM_FILE" and
                # the value element being a file path. This file path is to a JSON file that is loaded
                # and saved as subsettings. Subsettings is a dictionary of settings that is then passed
                # to the builder setup file which returns parameters saved under the original key. This
                # value element is now an object created by the builder instead of a dictionary
                subsettings = load_json(subsettings[BUILD_FROM_FILE])
                parameters[key] = builders[key](subsettings) if key in builders else key_error(key) 
            # if "FROM_FILE", read in JSON and save the new dictionary under the key
            elif subkey == FROM_FILE:
                # here subsettings is a dictionary of length=1 with only a key being "FROM_FILE" and
                # the value element being a file path. This file path is to a JSON file that is loaded
                # and saved as subsettings. Subsettings is a dictionary of settings that This
                # value element is now a new dictionary of subsettings loaded in from the JSON file.
                # The original lenght of 1 dictionary is now a new dictionary of new subsettings
                subsettings = load_json(subsettings[FROM_FILE])
                subsettings = load_build_and_get_parameters(settings=subsettings,builders=builders,*args, **kwargs)
                parameters[key] = subsettings
            # if it doesnt match any expression, just keep parameters[key] = settings[key] and move on
            else:
                pass
        elif isdict is True:
            subsettings = load_build_and_get_parameters(settings=subsettings,builders=builders,*args, **kwargs)
            parameters[key] = subsettings
                


    return parameters


class FailedPlasmaClassBuild(Exception):
    """Raised when a Plasma class is tried to be made but something fails"""

    def __init__(self, msg="Failed to make Plasma class"):
        self.msg = msg
        Exception.__init__(self, self.msg)


import numpy as np

from pstl.utls import constants as c
###############################################################################
# Functions for calculating current (position arguments important for wrappers here) #######
# make these more robust by changing position arguments to keyword arguments via dictionary
def wrapper_cylinderical_function(voltage, V_s, KT_e, area, n0, m_i):
    """
    Wrapper designed to elimeate the KT_e variable from passing to the
    call to the cylinderical function
    """
    return cylinderical_function(voltage, area, n0, V_s, m_i)

def wrapper_spherical_function(voltage, V_s, KT_e, area, n0, m_i):
    return spherical_function(voltage, area, n0, V_s, m_i, KT_e)


def wrapper_planar_function(voltage, V_s, KT_e, area, n0, m_i):
    return planar_function(voltage, area, n0, V_s, m_i, KT_e)
###############################################################################

###############################################################################
# Solves for currents #########################################################
def cylinderical_function(voltage, area, n0, V_s, m_i):
    A = (c.e*n0*area/np.pi)*np.sqrt(2*c.e/m_i)

    with np.errstate(invalid="ignore"):
        sqrt_x = np.sqrt(V_s-voltage)

    current = -np.multiply(
        A,
        sqrt_x
    )

    return current


def spherical_function(voltage, area, n0, V_s, m_i, KT_e):
    return _spherical_and_planar_function(voltage, area, n0, V_s, m_i, KT_e)


def planar_function(voltage, area, n0, V_s, m_i, KT_e):
    return _spherical_and_planar_function(voltage, area, n0, V_s, m_i, KT_e)
###############################################################################


###############################################################################
# Spherical and Planar (generalized to one function) ##########################
def _spherical_and_planar_function(voltage, area, n0, V_s, m_i, KT_e):
    A = (c.e*n0*area)*np.sqrt(c.e*KT_e/(2*np.pi*m_i))

    current = -np.multiply(A, np.divide(V_s-voltage, KT_e))

    return current
###############################################################################



import tkinter as tk

from pstl.gui.langmuir.control import SingleProbeLangmuirResultsControlFrame
from pstl.gui.langmuir.control import ControlFunctionFit, ControlEntryLabelFrame


def main():
    app = tk.Tk()

    control = SingleProbeLangmuirResultsControlFrame(master=app)
    #control=ControlFunctionFit(master=app)
    #control=ControlEntryLabelFrame(lbl_txt="HELO",master=app)


    control.grid(row=0,column=0,sticky="NSEW")
    control.tkraise()

    app.mainloop()


if __name__ == "__main__":
    main()


import tkinter as tk
import argparse
import json
import traceback

import numpy as np
import pandas as pd
from matplotlib import style
from PIL import Image, ImageTk

from pstl.gui.langmuir.canvas import LinearSemilogyCanvas, LinearSemilogyDoubleCanvas, LinearSemilogyDoubleCanvasSingleProbeLangmuir
from pstl.gui.langmuir import LinearSemilogyDoubleCanvasSingleProbeLangmuir as Canvas
from pstl.gui.langmuir import LinearSemilogySingleLangmuirCombinedData as LSSLCD
from pstl.gui.langmuir import LinearSemilogySingleLangmuirCombinedData2 as LSSLCD2
from pstl.gui.langmuir import LSSLCD_setup
from pstl.gui.langmuir import LSSLCD2_setup

from pstl.gui.langmuir import SingleProbeLangmuirResultsFrame as Panel
from pstl.gui.langmuir import SingleProbeLangmuirResultsFrame
from pstl.gui.langmuir import gui_langmuir

from pstl.utls.plasmas import XenonPlasma, ArgonPlasma, NeonPlasma, KryptonPlasma, Plasma
from pstl.utls.plasmas import setup as plasma_setup
from pstl.utls.errors.plasmas import FailedPlasmaClassBuild
from pstl.utls.errors.langmuir import Flagged

from pstl.utls.data import setup as data_setup

from pstl.extras.images import pstl_png_path, pstl_ico_path, pstl_16_ico_path, pstl_32_ico_path

from pstl.diagnostics.probes.langmuir.single import SphericalSingleProbeLangmuir as SSPL
from pstl.diagnostics.probes.langmuir.single import CylindericalSingleProbeLangmuir as CSPL
from pstl.diagnostics.probes.langmuir.single import PlanarSingleProbeLangmuir as PSPL
from pstl.diagnostics.probes.langmuir.single import setup as probe_setup

from pstl.diagnostics.probes.langmuir.single.analysis.solvers.solvers import SingleLangmuirProbeSolver as SLPS
from pstl.diagnostics.probes.langmuir.single.analysis.solvers.solvers import SingleLangmuirProbeSolver as SLPS

#style.use("bmh")

parser = argparse.ArgumentParser(
                    prog='GUI Langmuir',
                    description='Anaylsis on IV-trace',
                    epilog='Text at the bottom of help')
parser.add_argument('-o','--sname', help="save name for plots", default="outpng.png")
parser.add_argument('-f','--fname', help="file name to read in", default="lang_data.txt")
parser.add_argument('-c','--convergence', help="convergence rmse threshold for electron retarding fit", default=25,type=float)
parser.add_argument('-n','--negative', help="if electron current is negative",action="store_true")
parser.add_argument('-s','--delimiter', help="sets sep or delimiter of tabular data file, default is ',' use '\t' for tab",default=",",type=str)
parser.add_argument('-g','--neutral_gas', help="define gas composition i.e. Xenon, Argon, Neon",default="",type=str)
parser.add_argument('-p','--probe', help="define probe shape diameter length* where *only for spherical",nargs="*")

#parser.add_argument('-O','--output_settings_file', help="location of output results defining json file",type=str)
parser.add_argument('-G',   '--neutral_gas_settings_file',  help="location of neutral gas (and plasma) defining json file",  type=str)
parser.add_argument('-P',   '--probe_settings_file',        help="location of probe defining json file",                     type=str)
parser.add_argument('-D',   '--data_settings_file',         help="location of data defining json file",                      type=str)
parser.add_argument('-L',   '--solver_settings_file',       help="location of solver settings json file",                    type=str)

parser.add_argument('-S',   '--settings_file',              help="location of whole gui settings json file",                 type=str)
args = parser.parse_args()



available_shapes = {
    0: ['cylinderical', 'cylindrical'],
    1: ['spherical'],
    2: ['planar', 'planer'],
}

def old_main():
    x = np.linspace(-.2, .2, 9)
    y = x*.2

    root = tk.Tk()
    a = LinearSemilogyDoubleCanvasSingleProbeLangmuir(root, width=6, height=5)
    results = None
    b = SingleProbeLangmuirResultsFrame(root, results=results)
    a.add_raw_data(x, y)
    a.add_floating_potential(.2, color="r")
    a.add_electron_retarding_fill(-0.1, .1, color="black", alpha=0.1)
    a.legend()
    # a = LinearSemilogyCanvas(root)
    # a.widgets.frames['semilogy'].ax.relim()
    # a.widgets.frames['semilogy'].ax.autoscale()
    # a.widgets.frames['semilogy'].set_fig_to_semilogy()
    # a.ax2.update_from(a.ax1)
    # a.widgets.frames['semilogy'].fig.canvas.draw()
    # a.widgets.frames['linear'].fig.canvas.draw()

    # pack
    a.grid(row=0, column=1)
    b.grid(row=0, column=0)

    # run loop
    root.mainloop()


def get_lang_data():
    filename = args.fname
    data = pd.read_csv(filename, names=["voltage", "current"],header=1,delimiter=args.delimiter)
    if args.negative:
        data.iloc[:, 1] *= -1
    else:
        data.iloc[:, 1] *= 1
    return data

def get_settings():
    filename = args.sname # Change this
    thershold_rmse = args.convergence
    # filename (no extentsion) to open
    fname_split = args.fname.split(".")

    # canvas kwargs
    ptype = "png"
    sname = fname_split[0]+"_plot."+ptype
    canvas_kwargs = {
        'saveas': sname,
        'stype': ptype,
        'width': 5,
        'height': 4,
    }

    # panel kwargs
    ftype = "csv"
    fname = fname_split[0]+"_results."+ftype
    panel_kwargs = {
        'fname': fname,
        'ftype': ftype,
    }

    # solver kwargs
    # solver = SLPS(plasma, probe, data)
    solver_kwargs = {
    }


    settings = {
        'canvas_kwargs':canvas_kwargs,
        'panel_kwargs': panel_kwargs,
        'solver_kwargs': solver_kwargs,
    }
    return settings

"""
WLP: Cylinderical Downstream
Diameter: 0.1230 inches -> 0.31242 cm
Length: 0.6105 inches -> 1.55067 cm

PLP: Planer Down stream
Diameter: 0.1570 inches -> 0.39878 cm

NLP: Cylinderial @ Neutral
Diameter: 0.0155 inches -> 0.03937 cm
Lenth: 0.1170 inches -> 0.29718 cm

Farday:
Diameter inner: 0.6925 inches -> 1.175895 cm
Diameter outer: 1.120 inches -> 2.8448 cm
"""
def wlp():
    return CSPL(0.31242e-2,1.55067e-02)
def plp():
    return PSPL(0.39878e-02)
def NLP():
    return CSPL(0.03937e-02,0.29718e-02)
def blp():
    return SSPL(0.010, 0.0079)
def blp_smaller():
    return SSPL(0.010, 0.009) # BLP6
def rox():
    return CSPL(0.76e-3,2.54e-3)

def set_plasma_type(string,*args, **kwargs):
    plasma = None
    while plasma is None:
        plasma = logic_plasma_type(string, *args, **kwargs)
        if plasma is not None:
            break
        else:
            plasma = choose_plasma_type()
            
    return plasma

    
def choose_probe_type():
    shape = args.probe[0]
    shape = shape.lower()
    if shape in available_shapes[0]: #"cylinderical"
        Probe = CSPL
    elif shape in available_shapes[1]:  #"spherical":
        Probe = SSPL
    elif shape in available_shapes[2]:  #"planer"
        Probe = PSPL
    else:
        raise ValueError("First argument of probe does not match available shapes '%s'"%(shape))
    other_args = args.probe[1:]
    new_args:list[None|int|float] = [None]*len(other_args)
    for k, value in enumerate(other_args):
        try:
            new_args[k] = float(value)  
        except:
            raise ValueError("Argument %i of Probe is not a float %s"%(k+1,other_args[k]))
    tuple_args: tuple[float] = tuple(new_args)  # type: ignore

    probe = Probe(*tuple_args) # type: ignore
    return probe

    

def choose_plasma_type():
    string = input("Please enter Xenon, Argon, Neon\nor define using '-p' flag when initializing>>")
    plasma = logic_plasma_type(string)
    return plasma

def logic_plasma_type(string, *args, **kwargs):
    string = string.upper()
    if string == "":
        plasma = choose_plasma_type()
    elif string == "XENON":
        plasma = XenonPlasma(*args, **kwargs)
    elif string == "ARGON":
        plasma = ArgonPlasma(*args, **kwargs)
    elif string == "NEON":
        plasma = NeonPlasma(*args, **kwargs)
    else:
        try:
            plasma = Plasma(*args,**kwargs)
        except:
            print("'%s' does not match known options."%(string))
            raise FailedPlasmaClassBuild()

    return plasma

def gui_langmuir_from_file(settings_file:str):
    # initiate app
    app = tk.Tk()
    app.title("PSTL GUI Langmuir")

    with open(settings_file) as f: 
        settings = json.load(f)
    page = LSSLCD2_setup(settings, master=app)

    # pack it on
    page.pack()

    # close button
    btn = tk.Button(app,text="Close App",command=app.destroy)
    btn.pack()

    # run loop
    app.mainloop()
    



def main():
    if args.settings_file is None:
        if args.probe_settings_file is None and args.probe is None:
            raise ValueError("either probe_settings_file or probe must be defined")

        if args.neutral_gas_settings_file is None and args.plasma is None:
            raise ValueError("either neutral_gas_settings_file or plasma must be defined")
        
        if args.data_settings_file is None and args.fname is None:
            raise ValueError("either data_settings_file or fname must be defined")

    
    # load settings
    if args.settings_file is None:
        settings = get_settings()

        # probe settings
        if args.probe_settings_file is None: 
            probe = choose_probe_type()         
        else:
            with open(args.probe_settings_file) as f: 
                probe_settings = json.load(f)
            probe = probe_setup(probe_settings)

        # plasma settings
        if args.neutral_gas_settings_file is None: 
            plasma = choose_plasma_type() if args.gas is None else set_plasma_type(args.gas)
        else:
            with open(args.neutral_gas_settings_file) as f: 
                plasma_settings = json.load(f)
            plasma = plasma_setup(plasma_settings)
        
        # data settings
        if args.data_settings_file is None: 
            try:
                data = pd.read_csv(args.fname)
            except:
                raise ValueError("fname unable to load, must specify data_file or fix fname")
        else:
            with open(args.data_settings_file) as f:
                data_settings = json.load(f)
            data = data_setup(data_settings)
        # if negative
        if args.negative:
            data.iloc[:, 1] *= -1
        else:
            data.iloc[:, 1] *= 1

        # solver settings
        if args.solver_settings_file is None: 
            # solver args
            solver_args = {
                #'plasma': XenonPlasma(),
                'plasma': plasma,
                #'probe': SSPL(0.010, 0.0079),
                'probe': probe,
                'data': data,                # use -f <file-path\file-name.csv>
                }
            # create page
            # create combined langmuir frame that sits on one page
            print(solver_args,"\n",settings)
            #page = LSSLCD(**solver_args,master=app,**settings)
        else:
            with open(args.solver_settings_file) as f:
                solver_settings = json.load(f)
            solver = data_setup(solver_settings)
            settings["solver"] = solver
            #page = LSSLCD2_setup(settings, master=app)

    else:
        print("All settings present")
        with open(args.settings_file) as f: 
            settings = json.load(f)
    try:
        gui_langmuir(settings)
    except Flagged as e:
        print(e)

def olde_main(settings, app):
    # initiate app
    app = tk.Tk()
    app.title("PSTL GUI Langmuir")
    #pstl_png = tk.PhotoImage(file=pstl_png_path)
    app.iconphoto(
        True, 
        ImageTk.PhotoImage(Image.open(pstl_32_ico_path)),
        ImageTk.PhotoImage(Image.open(pstl_16_ico_path))
    )

    # Indented 1
    try:
        page = LSSLCD2_setup(settings, master=app)
    except:
        traceback.print_exc()
        file = settings["name"]+".tab"
        print("\nFAILED: ",file,"\n")
        #with open("/home/tyjoto/janus/temp/fail.tab") as f:
        #    f.write(file+"\n")

        pass
    else:
        def save_n_close():
            page.save_n_close()
            app.destroy()
        # save and close
        btn_savenclose = tk.Button(app,text="Save and Close app", command=save_n_close)
        btn_savenclose.pack()
    finally:
        # close button
        btn = tk.Button(app,text="Close App",command=app.destroy)
        btn.pack()

        # flag button
        btn_flag = tk.Button(app,text="Flag",command=lambda: print(repr("Flagged: {0}".format(str(settings.get("name","Unknown"))))))
        btn_flag.pack()

        # run loop
        app.mainloop()




def old_main_2():
    # initiate app
    app = tk.Tk()

    # create page
    page = tk.Frame(app)
    page.pack()
    
    # filename (no extentsion) to open
    fname_split = args.fname.split(".")

    # create Canvas
    ptype = "png"
    sname = fname_split[0]+"_plot."+ptype
    #canvas = Canvas(page, saveas=args.sname, width=4, height=3)
    canvas = Canvas(page, saveas=sname, width=5, height=4)
    canvas.grid(row=0, column=1, sticky="NSWE", rowspan=2)

    # create Panel
    ftype = "csv"
    fname = fname_split[0]+"_results."+ftype
    panel = Panel(page,fname=fname,ftype=ftype)
    panel.grid(row=0, column=0, sticky="N")

    data = get_lang_data()
    # data = get_test_data()

    # probe = CSPL(2.4e-3, 10e-3)
    probe = SSPL(0.010, 0.0079)
    probe_smaller = SSPL(0.010, 0.009) # probe 5 (plasma screen)
    rox_probe = CSPL(0.76e-3,2.54e-3)
    plasma = XenonPlasma()
    # plasma = ArgonPlasma()

    solver = SLPS(plasma, probe, data)

    # pre process data
    solver.preprocess()

    # solve for data
    solver.solve()

    # make plots of solved for plasma parameters
    canvas.make_plot(solver)

    # pass data to results panel
    panel.results = solver.results

    # run mainloop
    app.mainloop()


if __name__ == "__main__":
    main()



default_fit_thin_kwargs = {
    'deg': 1, 'power': 1, 'polarity': 1,
    'reverse': False, 'return_best': True, 'fit_type': "linear",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.30,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 1.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}


default_fit_power_kwargs = {
    'deg': 1, 'power': None, 'polarity': -1,  # power needs to be defined
    'reverse': False, 'return_best': True, 'fit_type': "power",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.50,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 10.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}

default_fit_kwargs = {
    "thin"  :   default_fit_thin_kwargs,
    "thick" :   default_fit_power_kwargs,
}


import numpy as np

from .shape_domains import (
    cylinderical_domain_condition,
    spherical_domain_condition,
    planar_domain_condition,
)
from .shape_func_helpers import(
    _function, _xs
)

def cylinderical_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p=None, lambda_D=None, *args, correct_area=True, **kwargs):
    if correct_area is True:
        if r_p is None or lambda_D is None:
            raise ValueError("'r_p' and 'lambda_D' must be a float: " +
                             str(type(r_p))+" and "+str(type(lambda_D)))
        xs = _xs(voltage, V_s, KT_e, lambda_D)
        area = area*(1+xs/r_p)
    else:
        area = np.ones_like(voltage)*area

    # condition
    condition = cylinderical_domain_condition(voltage, V_s, KT_e)

    # I_i function
    func = _function

    # get I_i
    I_i = np.where(
        condition,
        func(area, n0, m_i, KT_e),
        0,
    )

    # make returns
    extras = {"method": "thin", "shape": "cylinderical"}
    return I_i, extras


def spherical_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p=None, lambda_D=None, *args, correct_area=True, **kwargs):
    if correct_area is True:
        if r_p is None or lambda_D is None:
            raise ValueError("'r_p' and 'lambda_D' must be a float: " +
                             str(type(r_p))+" and "+str(type(lambda_D)))
        xs = _xs(voltage, V_s, KT_e, lambda_D)
        area = area*np.power(1+xs/r_p, 2)
    else:
        area = np.ones_like(voltage)*area

    # condition
    condition = spherical_domain_condition(voltage, V_s, KT_e)

    # I_i function
    func = _function

    # get I_i
    I_i = np.where(
        condition,
        func(area, n0, m_i, KT_e),
        0,
    )

    # make returns
    extras = {"method": "thin", "shape": "spherical"}
    return I_i, extras


def planar_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p=None, lambda_D=None, *args, correct_area=True, **kwargs):
    if correct_area is True:
        pass  # no correction if oriented correctly
        if r_p is None or lambda_D is None:
            # not needed because no correction (filler for later if orientation bad)
            pass
        area = np.ones_like(voltage)*area
    else:
        area = np.ones_like(voltage)*area

    # condition
    condition = planar_domain_condition(voltage, V_s, KT_e)

    # I_i function
    func = _function

    # get I_i
    I_i = np.where(
        condition,
        func(area, n0, m_i, KT_e),
        0,
    )

    # make returns
    extras = {"method": "thin", "shape": "planar"}
    return I_i, extras


from abc import ABC, abstractmethod

import numpy as np

class Probe(ABC):
    def __init__(self,name=None, description=None) -> None:
        super().__init__()
        self._name = name
        self._description = description

    @property
    def name(self):
        return self._name
    @name.setter
    def name(self, string):
        if isinstance(string, (str, type(None))):
            self._name = string
        else:
            raise TypeError("Name change must be str or None type, not type '%s'"%(str(type(string))))

    @property
    def description(self):
        return self._description
    @description.setter
    def description(self, string):
        if isinstance(string, (str, type(None))):
            self._description = string
        else:
            raise TypeError("Description change must be str or None type, not type '%s'"%(str(type(string))))
    